#$Id$
package XML::Storable;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $newline);
use Data::Dumper;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	xml_freeze
	xml_thaw
);
@EXPORT_OK=qw(
    xml_freeze
	xml_thaw
);
$VERSION = '0.01';

#######################################################################
sub freeze
{
    my $ret='';
    foreach my $e (@_)
    {
        $ret.=_freeze($e);
    }
    return $ret;
}
*xml_freeze=\&freeze;

#######################################################################
sub _freeze
{
    my $e=shift;
    return "<UNDEF></UNDEF>" if not defined $e;
    my $n='';
    $n="\n" if $newline;
    my $type=ref $e;
    my $ret;
    
    ## Blessed objects need to be gigamipokered...
    my $class='';
    if($e=~/^$type=(HASH|ARRAY|SCALAR)\(0x[0-9A-Fa-f]+\)$/)
    {
        return $e->xml_freeze() if $e->can('xml_freeze');
        $class=qq( class="$type");
        $type=$1;
    }

    if(not $type)
    {
        $ret='<S>'._xml_safe($e).'</S>'.$n;
    } elsif($type eq 'HASH')
    {
        $ret="<HASH$class>";
        while(my($k, $v)=each %$e)
        {
            $ret.='<KEY>'._xml_safe($k).'</KEY>'.$n.'<VALUE>'._freeze($v).'</VALUE>'.$n;
        }
        $ret.='</HASH>'.$n;
    } elsif($type eq 'ARRAY')
    {
        local $_;
        $ret="<ARRAY$class>";
        foreach (@$e)
        {
            $ret.=_freeze($_);
        }
        $ret.='</ARRAY>'.$n;    
    } elsif($type eq 'SCALAR')
    {
        $ret="<S ref=1$class>"._xml_safe($e).'</S>'.$n;  # What a kludge
    } else
    {
        die "Unknown type: $type\n";
    }
    return $ret;
}

#######################################################################
sub thaw
{
    my($text)=@_;
    study $text;        # Gives us 16% win

    my $hash=0;
    my($type, @accume, @accume, $att, $body, @stack, $end, $q, $pos);
    while($text=~m(<(/)?                            # possible end
                    (ARRAY|HASH|S|UNDEF|KEY|VALUE)  # tag
                    ([^>]*?)>                       # attributes
                    ([^<]*))gix)                    # and text-body
    {
        $end=$1; $type=$2; $att=$3; $body=$4;        
        if($end)            # END of a tag
        {
            # warn Dumper \@stack;
            ($q, $att, $body, $pos, $hash)=@{pop @stack};
            warn "<$q> ended with </$type>\n" if($q ne $type);
            $type=$q;
            if($type eq 'KEY')
            {
                if($hash==0)
                {
                    warn "<KEY> tag not in <HASH>\n";
                } elsif($hash==2)
                {
                    warn "<KEY> found, expecting <VALUE>";
                }
                $hash=2;
                $type='S';      # treat the KEY as a scalar
            } 
            if($type eq 'VALUE')
            {
                if($hash==0)
                {
                    warn "<KEY> tag not in <HASH>\n";
                } elsif($hash==1)
                {
                    warn "<VALUE> found, expecting <KEY>\n";
                }
                $hash=1;
                
                $body=pop @accume;          
                if($pos != @accume)
                {
                    warn "<VALUE> contained more then one element\n";
                    splice @accume, $pos;
                }
            } elsif($type eq 'S')
            {
                $body=_xml_unsafe($body);
                $body=\$body if($att=~/\bref=1\b/i);
            } elsif($type eq 'HASH')
            {
                if($hash==2)
                {
                    warn "Missing <VALUE>\n";
                    $#accume++;
                }
                $body={splice @accume, $pos};
                
            } elsif($type eq 'ARRAY')
            {
                $body=[splice @accume, $pos];
    
            } elsif($type eq 'UNDEF')
            {
                undef $body;
            }
            bless $body, $1 if $att=~m(\bclass=("?[^"]+"))i;
            push @accume, $body;
        } else
        {
            push @stack, [$type, $att, $body, scalar(@accume), $hash];
            $hash=($type eq 'HASH');
        }
    }
    return @accume if wantarray;
    return $accume[0] if 1==scalar @accume;
    return \@accume;
}
*xml_thaw=\&thaw;

#######################################################################
sub _xml_safe
{
    my($val)=@_;
    $val=~s(&)(&#38;)g;     # MUST BE FIRST
    $val=~s(/)(&#47;)g;
    $val=~s(<)(&#60;)g;
    $val=~s(>)(&#62;)g; 
    $val=~s(([\x00-\x1f]))('&#'.ord($1).';')ge; 
    return $val;
}

#######################################################################
sub _xml_unsafe
{
    my($val)=@_;
    $val=~s(&#(\d+);)(chr($1))ge; 
    $val=~s(&#38;)(&)g;     # MUST BE LAST
    return $val;
}


1;
__END__
#######################################################################
$Log$


=head1 NAME

XML::Storable - Create a bare-bones XML representation of complex data structures

=head1 SYNOPSIS

    use XML::Storable;
    my $stored=xml_freeze(\%table, \@chair);
    ... send it over a socket or something ...
    my ($table, $chair)=xml_thaw($stored);

=head1 DESCRIPTION


=head1 BUGS

It is currently very slow, compared to Storable.   ./benchmark for detail.  
But then Storable uses an XS to get it's speed.

The XML parser is hand coded and not very good at error handling.  XML is
expected to be very rigourously clean.

=head1 AUTHOR

Philip Gwyn, fil@pied.nu

=head1 SEE ALSO


=cut
